#include "prjlibs-include/standards.h"
#include <sys/types.h>
#include <termios.h>
#include <unistd.h>
#include <grp.h>
#include <signal.h>
#include <errno.h>

#include "sys_param.h"

#include "skalibs/include/stddjb.h"
#include "prjlibs-include/constants.h"
#include "constants.h"

char const* PROG="setstate";

static char const t_uid[]="uid_t";
static char const t_gid[]="gid_t";

int main(int argc, char** argv) {
  char const* x;
  unsigned int i;
  unsigned long l;
  int do_chroot=0;
  int do_leader=0;
  int do_noleader=0;
  int do_setsid=0;
  int do_tty_chown=0;
  int do_tty_fg=0;
  int do_setgid=0;
  int do_setgroups=0;
  int do_setuid=0;
  int do_home=0;

  if (argc<3)
    strerr_die3x(100, usage, PROG, " -rlLsofgGuh command [arg ...]");

  x=argv[1];
  if (*x=='-') ++x;
  while (*x!='\0') {
    switch (*x) {
      case 'r': do_chroot=1;    break;
      case 'l': do_leader=1;    break;
      case 'L': do_noleader=1;  break;
      case 's': do_setsid=1;    break;
      case 'o': do_tty_chown=1; break;
      case 'f': do_tty_fg=1;    break;
      case 'g': do_setgid=1;    break;
      case 'G': do_setgroups=1; break;
      case 'u': do_setuid=1;    break;
      case 'h': do_home=1;      break;
      default: {
        char opt[2]={ '\0', '\0' };
        *opt=*x;
        strerr_die3x(100, PROG, "unrecognized option: ", opt);
      }
    }
    ++x;
  }

  if (do_chroot) {
    x=env_get(var_root+1);
    if (x==null)
      strerr_die4x(100, PROG, ": ", var_root, err_notset);
    if (chdir(x)!=0)
      strerr_die5sys(errstat, PROG, err_unable, err_chdirto, x, ": ");
    if (chroot(".")!=0)
      strerr_die5sys(errstat, PROG, err_unable, err_chrootto, x, ": ");
  }

  if (do_leader)
    if (setpgid(0, 0)!=0)
      strerr_die4sys(100, PROG, err_unable, err_pgid, ": ");

  if (do_noleader) {
    pid_t pid;
    int pipefd[2];
    if (pipe(pipefd)==-1)
      strerr_die5sys(errstat, PROG, err_unable, err_create, err_pipe, ": ");
    pid=fork();
    if (pid==-1)
      strerr_die4sys(111, PROG, err_unable, err_fork, ": ");
    if (pid==0) {
      close(pipefd[0]);
      if (setpgid(0, 0)!=0)
        strerr_die4sys(100, PROG, err_unable, err_pgid, " in child: ");
      close(pipefd[1]);
      pause();
      _exit(0);
    }
    close(pipefd[1]);
    read(pipefd[0], &pipefd[1], 1);
    close(pipefd[0]);
    if (setpgid(getpid(), pid)!=0)
      strerr_die4sys(100, PROG, err_unable, err_pgid, ": ");
    kill(pid, SIGKILL);
    if (wait_pid(null, pid)!=pid)
      strerr_die4sys(100, PROG, err_unable, err_wait, ": ");
  }

  if (do_setsid)
    if (setsid()<0)
      strerr_die4sys(100, PROG, err_unable, err_sid, ": ");

  if (do_tty_fg) {
    sigset_t ttou, old;
    if (sigemptyset(&ttou)!=0)
      strerr_die5sys(100, PROG, err_unable, err_empty, err_sigset, ": ");
    if (sigaddset(&ttou, SIGTTOU)!=0)
      strerr_die5sys(100, PROG, err_unable, err_addto, err_sigset, ": ");
    if (sigprocmask(SIG_BLOCK, &ttou, &old)!=0)
      strerr_die4sys(100, PROG, err_unable, err_sigmask, ": ");
    if (tcsetpgrp(0, getpgid(0)))
      strerr_die3sys(100, PROG, err_unable,
                     "set foreground process group id: ");
    if (sigprocmask(SIG_SETMASK, &old, null)!=0)
      strerr_die4sys(100, PROG, err_unable, err_sigmask, ": ");
  }

  if (do_tty_chown) {
    x=env_get(var_uid+1);
    if (x==null)
      strerr_die4x(100, PROG, ": ", var_uid, err_notset);
    i=scan_ulong(x, &l);
    if (i==0 || x[i]!='\0')
      strerr_die4x(100, PROG, ": ", var_uid, err_malformed);
    if (l!=(uid_t)l)
      strerr_die5x(100, PROG, ": ", var_uid, err_oflow, t_uid);
    if (fchown(fileno_stdin, l, -1)!=0)
      strerr_die4sys(errstat, PROG, err_unable, err_chown, "fd 0: ");
  }

  if (do_setgid) {
    x=env_get(var_gid+1);
    if (x==null)
      strerr_die4x(100, PROG, ": ", var_gid, err_notset);
    i=scan_ulong(x, &l);
    if (i==0 || x[i]!='\0')
      strerr_die4x(100, PROG, ": ", var_gid, err_malformed);
    if (l!=(gid_t)l)
      strerr_die5x(100, PROG, ": ", var_gid, err_oflow, t_gid);
    if (setgid(l)!=0)
      strerr_die5sys(100, PROG, err_unable, err_setgidto, x, ": ");
  }

  if (do_setgroups) {
    size_t ngroups;
    gid_t* groups;
    static char const toolong[]="too long";
    x=env_get(var_groups+1);
    if (x==null)
      strerr_die4x(100, PROG, ": ", var_groups, err_notset);
    {
      char const* y=x;
      ngroups=0;
      while (*y==' ') {
        ++ngroups;
        if (ngroups==0)
          strerr_die4x(100, PROG, ": ", var_groups, toolong);
        ++y;
        y+=str_chr(y, ' ');
      }
    }
    i=ngroups*sizeof (gid_t);
    if (i/sizeof (gid_t)!=ngroups)
      strerr_die4x(100, PROG, ": ", var_groups, toolong);
    groups=(gid_t*)alloc(i);
    if (groups==null)
      strerr_die4sys(111, PROG, err_unable, err_alloc, ": ");
    ngroups=0;
    while (*x!='\0') {
      gid_t gid;
      if (*x!=' ')
        strerr_die4x(100, PROG, ": ", var_groups, err_malformed);
      ++x;
      i=scan_ulong(x, &l);
      x+=i;
      if (i==0)
        strerr_die4x(100, PROG, ": ", var_groups, err_malformed);
      gid=l;
      if (gid!=l)
        strerr_die5x(100, PROG, ": group id in ", var_groups, err_oflow,
                     t_gid);
      groups[ngroups]=gid;
      ++ngroups;
    }
    if (setgroups(ngroups, groups)!=0)
      strerr_die3sys(100, PROG, err_unable, "set supplementary groups: ");
  }

  if (do_setuid) {
    x=env_get(var_uid+1);
    if (x==null)
      strerr_die4x(100, PROG, ": ", var_uid, err_notset);
    i=scan_ulong(x, &l);
    if (i==0 || x[i]!='\0')
      strerr_die4x(100, PROG, ": ", var_uid, err_malformed);
    if (l!=(uid_t)l)
      strerr_die5x(100, PROG, ": ", var_uid , err_oflow, t_uid);
    if (setuid(l)!=0)
      strerr_die5sys(100, PROG, err_unable, err_setuidto, x, ": ");
  }

  if (do_home) {
    x=env_get(var_home+1);
    if (x==null)
      strerr_die4x(100, PROG, ": ", var_home, err_notset);
    if (chdir(x)!=0)
      strerr_die5sys(errstat, PROG, err_unable, err_chdirto, x, ": ");
  }

  argv+=2;
  pathexec_run(argv[0], (char const**)argv, (char const**)environ);
  strerr_die5sys(errstat, PROG, err_unable, err_exec, argv[0], ": ");
}
